#include <iostream>
#include <fstream>

#include <torch/ConnectedMachine.h>
#include <torch/MLP.h>
#include <torch/Linear.h>

#include <torch/MatDataSet.h>
#include <torch/TwoClassFormat.h>
#include <torch/OneHotClassFormat.h>
#include <torch/DiskXFile.h>
#include <torch/MemoryXFile.h>

#include <torch/ClassMeasurer.h>
#include <torch/MSECriterion.h>
#include <torch/MSEMeasurer.h>
#include <torch/StochasticGradient.h>

#include <torch/IOAscii.h>

using namespace Torch;
using namespace std;

int main(int argc, char** argv) {
	int n_inputs = 7;
	int n_hidden = 7;
	int n_targets = 7;

	Allocator* allocator = new Allocator;

//	ConnectedMachine mlp;
//	Linear* l1 = new(allocator) Linear(n_inputs, n_hidden);
//	mlp.addFCL(l1);
//	Linear* l2 = new(allocator) Linear(n_hidden, n_targets);
//	mlp.addFCL(l2);
	
	// Creates a 2 -> 2 -> 1 Linear and Fully Connected MultiLayerPerceptron 
	MLP mlp(2, n_inputs, "linear", n_hidden, "linear", n_targets);
	mlp.build();
	mlp.setPartialBackprop(false);

	char traindatafile[] = "data/KA_train.gz";

	DataSet* data = new(allocator) 
		MatDataSet(traindatafile, n_inputs, n_targets);
	OneHotClassFormat* class_format = 
		new(allocator) OneHotClassFormat(data);
	
	// create 2 Measurer, one writing to stdout and one to file
	MeasurerList measurers; 
/*
//	 measurers.addNode(
//		new(allocator) ClassMeasurer(mlp.outputs, data, \
//				class_format, 
//				new(allocator) DiskXFile(stdout)));
*/	
	measurers.addNode(
		new(allocator) ClassMeasurer(mlp.outputs, data, \
				class_format, 
				new(allocator) DiskXFile("measurer.output", "w")));

	Criterion* criterion = new(allocator) MSECriterion(n_inputs);
	
	StochasticGradient trainer(&mlp, criterion);

	unsigned int iterations = 500;
	if(argc>1)
		iterations = atoi(argv[1]);
	trainer.setIOption("max iter", iterations);
		
	//trainer.setROption("end accuracy", 0.000001);
	//trainer.setROption("learning rate", 10);

	trainer.train(data, &measurers);

	// reload data files to feed and test outputs of net
	DiskXFile* out = new(allocator) DiskXFile("cout", "w");

	// and now feed MLP with read data
	DataSet* testdata = data;

	cout << "ciao\n";
	out->printf("testdata->n_inputs n_examples, n_frames and frame_size: %d %d %d %d\n", 
		testdata->n_examples,
		testdata->n_inputs,
		testdata->inputs->n_frames,
		testdata->inputs->frame_size);
	cout << "ciao\n";
	out->printf("mlp.outputs->n_frames and frame_size: %d %d\n", \
		mlp.outputs->n_frames,
		mlp.outputs->frame_size);

	out->flush();
	cout << "ciao\n";
 
	// show next net input
	for(int example=0; example < testdata->n_examples; ++example) {
		testdata->setExample(example);
		testdata->pushExample();
		out->printf("SetExample %d\n", example);

		for(int nfi=0; nfi < testdata->inputs->n_frames; ++nfi) 
			for(int ni=0; ni < testdata->inputs->frame_size; ++ni) {
				out->printf("example %d, %X input[%d][%d]: %d\n",
					example, 
					&(testdata->inputs->frames[nfi][ni]),
					nfi, ni, 
					testdata->inputs->frames[nfi][ni]);
				
				out->printf("example %d, output[%d][%d]: %d\n",
					example, nfi, ni, 
					testdata->targets->frames[nfi][ni]);
			}

		out->flush();
		// feed the net
		mlp.forward(testdata->inputs);
		// and show the result
		for(int nfo=0; nfo < mlp.outputs->n_frames; ++nfo) 
			for(int no=0; no < mlp.outputs->frame_size; ++no)
				out->printf("example %d outputs: %d\n",
				example,
				mlp.outputs->frames[nfo][no]);
		out->printf("\n");
		testdata->popExample();
		out->flush();
	}

	out->printf("END\n");
	delete allocator;
	return 0;
}
