
#include <iostream>
#include <fstream>

#include <torch/ConnectedMachine.h>
#include <torch/MLP.h>
#include <torch/Linear.h>

#include <torch/MatDataSet.h>
#include <torch/TwoClassFormat.h>
#include <torch/OneHotClassFormat.h>
#include <torch/DiskXFile.h>
#include <torch/MemoryXFile.h>

#include <torch/ClassMeasurer.h>
#include <torch/MSECriterion.h>
#include <torch/MSEMeasurer.h>
#include <torch/StochasticGradient.h>

#include <torch/IOAscii.h>

using namespace Torch;
using namespace std;

int main(int argc, char** argv) {
	int n_inputs = 2;
	int n_hidden = 2;
	int n_targets = 1;

	Allocator* allocator = new Allocator;

	ConnectedMachine mlp;
	Linear* l1 = new(allocator) Linear(n_inputs, n_hidden);
	mlp.addFCL(l1);
	Linear* l2 = new(allocator) Linear(n_hidden, n_targets);
	mlp.addFCL(l2);
	
	// Creates a 2 -> 2 -> 1 Linear and Fully Connected MultiLayerPerceptron 
	//MLP mlp(2, 2, "linear", 2, "linear", 1);
	mlp.build();
	mlp.setPartialBackprop(false);

	
	char traindatafile[] = "data/xor_train.gz";


	DataSet* data = new(allocator) 
		MatDataSet(traindatafile, n_inputs, n_targets);
	OneHotClassFormat* class_format = new(allocator) OneHotClassFormat(data);
	
	// create 2 Measurer, one writing to stdout and one to file
	MeasurerList measurers; 
//	measurers.addNode(
//		new(allocator) ClassMeasurer(mlp.outputs, data, \
//				class_format, 
//				new(allocator) DiskXFile(stdout)));
	
	measurers.addNode(
		new(allocator) ClassMeasurer(mlp.outputs, data, \
				class_format, 
				new(allocator) DiskXFile("measurer.output", "w")));

	Criterion* criterion = new(allocator) MSECriterion(n_inputs);
	
	StochasticGradient trainer(&mlp, criterion);

	unsigned int iterations = 500;
	if(argc>1)
		iterations = atoi(argv[1]);
	trainer.setIOption("max iter", iterations);
		
	//trainer.setROption("end accuracy", 0.000001);
	trainer.setROption("learning rate", 10);

	trainer.train(data, &measurers);

	trainer.test(&measurers);
	// reload data files to feed and test outputs of net
	IOAscii* ioascii = new(allocator) IOAscii("data/xor_train.gz");
	cout << "N. of Sequences: " << ioascii->n_sequences << endl;

	cout << "mlp.n_inputs: " << mlp.n_inputs << endl;
	cout << "mlp.n_outputs: " << mlp.n_outputs << \
		", mlp.outputs.n_frames: " << mlp.outputs->n_frames << endl;

	// and now feed MLP with read data
	for(int i=0; i < ioascii->n_sequences; ++i) {
		cout << "iteration " << i << endl;
		
		Sequence* seq = \
			new(allocator) Sequence(ioascii->getNumberOfFrames(i),
				ioascii->frame_size);
		ioascii->getSequence(i, seq);
		
		// show next net input
		for(int ni=0; ni < mlp.n_inputs; ++ni) {
			cout << "input: " << seq->frames[0][ni] << endl;
		}
		// feed the net
		mlp.forward(seq);
		// and show the result
		for(int no=0; no < mlp.n_outputs; ++no) {
			cout << "outputs: " << mlp.outputs->frames[0][no] << endl;
		}
		cout << endl << endl;

		delete seq;
	}	

	cout << "END" << endl;


	delete allocator;
	return 0;
}
