#include <iostream>
#include <fstream>

#include <torch/ConnectedMachine.h>
#include <torch/MLP.h>
#include <torch/Linear.h>

#include <torch/MatDataSet.h>
#include <torch/MeanVarNorm.h>
#include <torch/MultiClassFormat.h>
#include <torch/TwoClassFormat.h>
#include <torch/OneHotClassFormat.h>
#include <torch/DiskXFile.h>
#include <torch/MemoryXFile.h>

#include <torch/ClassMeasurer.h>
#include <torch/MSECriterion.h>
#include <torch/MSEMeasurer.h>
#include <torch/StochasticGradient.h>

#include <torch/IOAscii.h>

using namespace Torch;
using namespace std;

int main(int argc, char** argv) {
	int n_inputs = 2;
	int n_hidden = 2;
	int n_targets = 1;

	Allocator* allocator = new Allocator;

	// reload data files to feed and test outputs of net
	DiskXFile out("cout", "w");


	// Creates a 2 -> 2 -> 1 Linear and Fully Connected MultiLayerPerceptron 
	MLP mlp(2, n_inputs, "linear", n_hidden, "linear", n_targets);
	mlp.build();
	mlp.setPartialBackprop(false);

	char traindatafile[] = "data/KA2_train";

	MatDataSet data(traindatafile, n_inputs, n_targets);
	MeanVarNorm* mvn = new(allocator) MeanVarNorm(&data);
	data.preProcess(mvn);
	
	//OneHotClassFormat* class_format = 
	//	new(allocator) OneHotClassFormat(&data);
	MultiClassFormat* class_format =
		new(allocator) MultiClassFormat(&data);

	// create 2 Measurer, one writing to stdout and one to file
	MeasurerList measurers; 
/*
//	 measurers.addNode(
//		new(allocator) ClassMeasurer(mlp.outputs, data, \
//				class_format, 
//				new(allocator) DiskXFile(stdout)));
*/	
	measurers.addNode(
		new(allocator) ClassMeasurer(mlp.outputs, &data, \
				class_format, 
				new(allocator) DiskXFile("measurer.output", "w")));

	Criterion* criterion = 
		new(allocator) MSECriterion(n_inputs);
	
	StochasticGradient trainer(&mlp, criterion);

	unsigned int iterations = 500;
	if(argc>1)
		iterations = atoi(argv[1]);
	trainer.setIOption("max iter", iterations);
		
	//trainer.setROption("end accuracy", 0.000001);
	trainer.setROption("learning rate", .5);

	trainer.train(&data, &measurers);

	// and now feed MLP with read data
	out.printf("example in DataSet: %d\n",
		data.n_examples);
	out.printf(
		"input: neurons, frames and frameSize: "
		"%d %d %d\n", 
		data.n_inputs,
		data.inputs->n_frames,
		data.inputs->frame_size);
	out.printf(
		"output: neurons, frames and frameSize: "
		"%d %d %d\n", 
		data.n_targets,
		data.targets->n_frames,
		data.targets->frame_size);
	out.printf(
		"mlp.outputs->n_frames and frame_size: %d %d\n", \
		mlp.outputs->n_frames,
		mlp.outputs->frame_size);

	out.flush();
 
	// show next net input
	for(int example=0; example < data.n_examples; ++example) {
		data.setExample(example);
		out.printf("SetExample %d\n", example);

		for(int nfi=0; nfi < data.inputs->n_frames; ++nfi) {
			out.flush();
			out.printf("inputs  @0x%X, ex %d, frame %d: ", 
				&(data.inputs->frames[nfi]),
				example, nfi);
			for(int ni=0; ni < data.inputs->frame_size; ++ni)
				out.printf("%f (%d), ",
					(data.inputs->frames)[nfi][ni], ni);
			out.flush();
		}
		out.printf("\n");

		for(int nfi=0; nfi < data.targets->n_frames; ++nfi) { 
			out.flush();
			out.printf("targets @0x%X, ex %d, frame %d: ", 
				&(data.targets->frames[nfi]),
				example, nfi);
			for(int ni=0; ni < data.targets->frame_size; ++ni)
				out.printf("%f (%d), ",
					(data.targets->frames)[nfi][ni], ni);
			out.flush();
		}
		out.printf("\n");
		out.flush();

		// feed the net
		mlp.forward(data.inputs);
		// and show the result
		for(int nfo=0; nfo < mlp.outputs->n_frames; ++nfo) { 
			out.printf("result @0x%X, ex %d, frame %d: ", 
				&(mlp.outputs->frames[nfo]), example, nfo);
			for(int no=0; no < mlp.outputs->frame_size; ++no)
				out.printf("%f (%d)\n",
					mlp.outputs->frames[nfo][no], no);
		}
		out.printf("\n");
		out.flush();
	}

	out.printf("END\n");
	delete allocator;
	return 0;
}
